(*******************************************************************************
*                                                                              *
*                             Delphi BasicObects                               *
*                                                                              *
*                     (C)Copyright 2001, Ferns Paanakker.                      *
*                                                                              *
*                                                                              *
*    This program is free software; you can redistribute it and/or modify      *
*    it under the terms of the GNU General Public License as published by      *
*    the Free Software Foundation; either version 2 of the License, or         *
*    (at your option) any later version.                                       *
*                                                                              *
*    This program is distributed in the hope that it will be useful,           *
*    but WITHOUT ANY WARRANTY; without even the implied warranty of            *
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
*    GNU General Public License for more details.                              *
*                                                                              *
*    You should have received a copy of the GNU General Public License         *
*    along with this program; if not, write to the Free Software               *
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
*******************************************************************************)

unit BasicObjects;

interface

uses classes, sysutils, math,
  BasicType, BasicSettings, BasicConst, BasicAssist;

type
  TVisualObject = class
  private
    FTimeCreate: integer;
    FTimeLastDecay: integer;
    FDecayRate: double;
    FConfidence: double;
    FKindSet: TKindSet;
    FRefCount: integer;
    FDirection: double;
    FDistance: double;
    FDistChng: double;
    FDirChng: double;
    FObjectPosition: TObjectPosition;

    function GetIsGoal: boolean;
    function GetIsPlayer: boolean;
    function GetIsBall: boolean;
    function GetIsRightGoal: boolean;
    function GetIsOfLeftTeam: boolean;
    function GetIsTheirGoal: boolean;
    function GetIsOwnGoal: boolean;
    function GetIsLeftGoal: boolean;
    function GetIsOfOwnTeam: boolean;
    function GetIsOfRightTeam: boolean;
    function GetIsFlag: boolean;
    function GetIsOfTheirTeam: boolean;
    function GetIsFixed: boolean;
    function GetIsLine: boolean;
    function GetIsOfUnknownTeam: boolean;
  public
    property ObjectPosition: TObjectPosition read FObjectPosition;

    property Direction: double read FDirection write FDirection;
    property Distance: double read FDistance write FDistance;
    property DistChng: double read FDistChng write FDistChng;
    property DirChng: double read FDirChng write FDirChng;

    property TimeCreate: integer read FTimeCreate;
    property DecayRate: double read FDecayRate;
    property Confidence: double read FConfidence;

    property isBall: boolean read GetIsBall;
    property isGoal: boolean read GetIsGoal;
    property isPlayer: boolean read GetIsPlayer;
    property isFlag: boolean read GetIsFlag;
    property isLine: boolean read GetIsLine;
    property isFixed: boolean read GetIsFixed;
    property isOwnGoal: boolean read GetIsOwnGoal;
    property isTheirGoal: boolean read GetIsTheirGoal;
    property isOfOwnTeam: boolean read GetIsOfOwnTeam;
    property isOfTheirTeam: boolean read GetIsOfTheirTeam;
    property isOfUnknownTeam: boolean read GetIsOfUnknownTeam;
    property isLeftGoal: boolean read GetIsLeftGoal;
    property isRightGoal: boolean read GetIsRightGoal;
    property isOfLeftTeam: boolean read GetIsOfLeftTeam;
    property isOfRightTeam: boolean read GetIsOfRightTeam;

    function isEqual(obj: TObject): boolean; virtual; abstract;
    function isOfKind(ks: TKindSet): boolean;
    procedure updateDecay(NewTime: integer);
    procedure SetFieldPosition(const X,Y: double);

    constructor Create(TimeCreate: integer);
  end;

  TVOList = class(TList)
  public
    function getObject(kind: TKind): TVisualObject;
  end;

  TVisualObjectList = class(TList)
  private
  public
    procedure Clear; override;
    function AddObject(obj: TVisualObject): Integer;
    function GetVOList(kindSet: TKindSet; minConfidence, maxConfidence: double):
      TVOList;
    procedure UpdateDecay(nTime: integer);
    destructor Destroy; override;
  end;

  TPlayerObject = class(TVisualObject)
  private
    FTeamName: string;
    FUniformNumber: integer;
    FBodyDir: double;
    FHeadDir: double;
    function GetFaceDir: double;
  public
    property TeamName : string read FTeamName;
    property BodyDir : double read FBodyDir write FBodyDir;
    property HeadDir : double read FHeadDir write FHeadDir;
    property UniformNumber : integer read FUniformNumber;
    property FaceDir: double read GetFaceDir;
    function isEqual(obj: TObject): boolean; override;
    constructor Create(TimeCreate: integer; ownTeamName: string; ownSide: TSide;
      TeamName: string; UniformNumber: integer);
  end;

  TImmovableObject = class(TVisualObject)
  private
    FID: string;
  public
    procedure setAbsPosition;
    constructor Create(TimeCreate: integer);
  end;

  TBallObject = class(TImmovableObject)
  private
  public
    function isEqual(obj: TObject): boolean; override;
    function isKickable: boolean;
    constructor Create(TimeCreate: integer);
  end;

  TFlagObject = class(TImmovableObject)
  private
    Fposition: set of TPosition;
    FinPenaltyArea: boolean;
    FisGoalpost: boolean;
  public
    property inPenaltyArea: boolean read FinPenaltyArea;
    property isGoalpost: boolean read FisGoalPost;
    function isEqual(obj: TObject): boolean; override;
    constructor Create(ID: string; TimeCreate: integer);
  end;

  TGoalObject = class(TImmovableObject)
  private
    FSide: TSide;
  public
    property Side : TSide read FSide;
    function isEqual(obj: TObject): boolean; override;
    constructor Create(TimeCreate: integer; ownSide: TSide; Side: TSide);
  end;

  TLineObject = class(TVisualObject)
  private
    FPosition: TPosition;
  public
    property Position: TPosition read FPosition;
    function isEqual(obj: TObject): boolean; override;
    constructor Create(TimeCreate: integer; position: TPosition);
  end;

implementation

{ TVisualObjectList }

(* Add object ot the visual object list. If the list already contains this
   object, then free the old object and insert the new object in it's place.
   Return the index where the object is added.
*)

function TVisualObjectList.AddObject(obj: TVisualObject): Integer;
var
  found: boolean;
  i: integer;
  visObj: TVisualObject;
begin
  found := false;
  for i := 0 to count - 1 do
  begin
    visObj := TVisualObject(items[i]);
    if visObj.isEqual(obj) then
    begin
      found := true;
      // update the information...remove the old information
      items[i] := obj;
      visObj.Free;
      result := i;
      break;
    end;
  end;

  if not found then
    result := add(pointer(obj));
end;

procedure TVisualObjectList.Clear;
var
  i: integer;
begin
  for i := 0 to Count - 1 do
    TVisualObject(items[i]).Free;
  inherited Clear;
end;

destructor TVisualObjectList.Destroy;
begin
  Clear;
  inherited Destroy;
end;

function TVisualObjectList.GetVOList(kindSet: TKindSet;
  minConfidence, maxConfidence: double): TVOList;
var
  i: integer;
  vo: TVisualObject;
begin
  Result := TVOList.Create;
  for i := 0 to Count - 1 do
  begin
    vo := items[i];
    if (vo.isOfKind(kindSet) and (vo.Confidence >= minConfidence) and
      (vo.Confidence <= maxConfidence)) then
    begin
      Result.Add(vo);
    end;
  end;
end;

procedure TVisualObjectList.UpdateDecay(nTime: integer);
var
  i: integer;
begin
  // Update the decau of all the objects
  for i := 0 to Count - 1 do
    TVisualObject(Items[i]).UpdateDecay(nTime);
end;

{ TFlagObject }

constructor TFlagObject.Create(ID: string; TimeCreate: integer);
begin
  inherited Create(TimeCreate);
  FKindSet  := [kFlagKnown];
  FID := lowercase(ID);
  FPosition := [];
  if (Pos('r',FID) > 0) then
    FPosition := FPosition + [poRight];
  if (Pos('l',FID) > 0) then
    FPosition := FPosition + [poLeft];
  if (Pos('t',FID) > 0) then
    FPosition := FPosition + [poTop];
  if (Pos('b',FID) > 0) then
    FPosition := FPosition + [poBottom];
  FinPenaltyArea := (Pos('p',FID) > 0);
  FisGoalpost := (Pos('g',FID) > 0);

  SetAbsPosition;
end;

function TFlagObject.isEqual(obj: TObject): boolean;
begin
  result := false;
  if obj is TFlagObject then
  begin
    if (TFlagObject(obj).FID = FID) then
    begin
      result := true;
    end;
  end;
end;

constructor TImmovableObject.Create(TimeCreate: integer);
begin
  inherited Create(TimeCreate);
  FDecayRate := DECAY_FIXED;
end;

procedure TImmovableObject.setAbsPosition;
var
  i: integer;
begin
  // Sets absolute position.
  for i := low(D_IMMOVABLE) to high(D_IMMOVABLE) do
  begin
    if FID = D_IMMOVABLE[i].ID then
    begin
      FObjectPosition := D_IMMOVABLE[i];
      break;
    end;
  end;
end;

{ TLineObject }

constructor TLineObject.Create(TimeCreate: integer; position: TPosition);
begin
  inherited Create(TimeCreate);
  FDecayRate := DECAY_FIXED;
  Fposition := position;
  FKindSet := [kLine];
end;

function TLineObject.isEqual(obj: TObject): boolean;
begin
  result := false;
  if obj is TLineObject then
    result := true;
end;

{ TBallObject }

constructor TBallObject.Create(TimeCreate: integer);
begin
  inherited Create(TimeCreate);
  FDecayRate := DECAY_BALL;
  FKindSet := [kBall];
end;

function TBallObject.isEqual(obj: TObject): boolean;
begin
  result := false;
  if obj is TBallObject then
    result := true;
end;

function TBallObject.isKickable: boolean;
begin
  result := (Distance < (PlaySettings.KickableMargin + PlaySettings.BallSize +
    PlaySettings.PlayerSize));
end;

{ TPlayerObject }

constructor TPlayerObject.Create(TimeCreate: integer; ownTeamName: string;
  ownSide: TSide; TeamName: string; UniformNumber: integer);
begin
  inherited Create(TimeCreate);
  FTeamName := TeamName;
  FUniformNumber := UniformNumber;

  if TeamName = ownTeamName then
    FKindSet := [kPlayerOwn]
  else
    FKindSet := [kPlayerOpponent];

  if ownSide = side_left then
    FKindSet := FKindSet + [kPlayerLeft]
  else
    FKindSet := FKindSet + [kPlayerRight];
end;

function TPlayerObject.GetFaceDir: double;
begin
  Result := FBodyDir + FHeadDir;
end;

function TPlayerObject.isEqual(obj: TObject): boolean;
begin
  result := false;
  if obj is TPlayerObject then
  begin
    if ((TPlayerObject(obj).TeamName = TeamName) and
      (TPlayerObject(obj).UniformNumber = UniformNumber)) then
    begin
      result := true;
    end;
  end;
end;

{ TGoalObject }

constructor TGoalObject.Create(TimeCreate: integer; ownSide: TSide; Side:
  TSide);
begin
  inherited Create(TimeCreate);
  FSide := Side;
  FID := SideToStr(Side);

  if ownSide = Side then
    FKindSet := [kGoalOwn]
  else
    FKindSet := [kGoalOpponent];

  if Side = side_left then
  begin
    FKindSet := FKindSet + [kGoalLeft];
  end
  else
  begin
    FKindSet := FKindSet + [kGoalRight];
  end;

  SetAbsPosition;
end;

function TGoalObject.isEqual(obj: TObject): boolean;
begin
  result := false;
  if obj is TGoalObject then
    result := TGoalObject(obj).Side = Side;
end;

{ TVisualObject }

constructor TVisualObject.Create(TimeCreate: integer);
begin
  FKindSet := [];
  FRefCount := 0;
  FDecayRate := DECAY_NORMAL;
  FTimeCreate := TimeCreate;
  FTimeLastDecay := FTimeCreate;
  FConfidence := 1.0;
end;

function TVisualObject.GetIsBall: boolean;
begin
  result := FKindSet * TKindBall <> [];
end;

function TVisualObject.GetIsFixed: boolean;
begin
  result := FKindSet * (TKindGoal + TKindFlag + TKindLine) <> [];
end;

function TVisualObject.GetIsFlag: boolean;
begin
  result := FKindSet * TKindFlag <> [];
end;

function TVisualObject.GetIsGoal: boolean;
begin
  result := FKindSet * TKindGoal <> [];
end;

function TVisualObject.GetIsLeftGoal: boolean;
begin
  result := FKindSet * [kGoalLeft] <> [];
end;

function TVisualObject.GetIsLine: boolean;
begin
  result := FKindSet * TKindLine <> [];
end;

function TVisualObject.GetIsOfLeftTeam: boolean;
begin
  result := FKindSet * [kPlayerLeft] <> [];
end;

function TVisualObject.GetIsOfOwnTeam: boolean;
begin
  result := FKindSet * [kPlayerOwn] <> [];
end;

function TVisualObject.GetIsOfRightTeam: boolean;
begin
  result := FKindSet * [kPlayerRight] <> [];
end;

function TVisualObject.GetIsOfTheirTeam: boolean;
begin
  result := FKindSet * [kPlayerOpponent] <> [];
end;

function TVisualObject.GetIsOfUnknownTeam: boolean;
begin
  result := FKindSet * [kPlayerUnknown] <> [];
end;

function TVisualObject.GetIsOwnGoal: boolean;
begin
  result := FKindSet * [kGoalOwn] <> [];
end;

function TVisualObject.GetIsPlayer: boolean;
begin
  result := FKindSet * TKindPlayer <> [];
end;

function TVisualObject.GetIsRightGoal: boolean;
begin
  result := FKindSet * [kGoalRight] <> [];
end;

function TVisualObject.GetIsTheirGoal: boolean;
begin
  result := FKindSet * [kGoalOpponent] <> [];
end;

function TVisualObject.isOfKind(ks: TKindSet): boolean;
begin
  result := (ks * FKindSet) <> [];
end;

procedure TVisualObject.SetFieldPosition(const X, Y: double);
begin
  FObjectPosition.Pos.X := X;
  FObjectPosition.Pos.Y := Y;
end;

procedure TVisualObject.updateDecay(NewTime: integer);
begin
  if NewTime > FTimeLastDecay then
  begin
    FConfidence := FConfidence * power(FDecayRate, (NewTime - FTimeLastDecay));
    FTimeLastDecay := NewTime;
  end;
end;

{ TVOList }

function TVOList.getObject(kind: TKind): TVisualObject;
var
  i: integer;
begin
  Result := nil;
  for i := 0 to Count - 1 do
    if TVisualObject(items[i]).isOfKind([kind]) then
    begin
      Result := items[i];
      break;
    end;
end;

end.
