(*******************************************************************************
*                                                                              *
*                             Delphi BasicPlayer                               *
*                                                                              *
*                     (C)Copyright 2001, Ferns Paanakker.                      *
*                                                                              *
*                                                                              *
*    This program is free software; you can redistribute it and/or modify      *
*    it under the terms of the GNU General Public License as published by      *
*    the Free Software Foundation; either version 2 of the License, or         *
*    (at your option) any later version.                                       *
*                                                                              *
*    This program is distributed in the hope that it will be useful,           *
*    but WITHOUT ANY WARRANTY; without even the implied warranty of            *
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
*    GNU General Public License for more details.                              *
*                                                                              *
*    You should have received a copy of the GNU General Public License         *
*    along with this program; if not, write to the Free Software               *
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
*******************************************************************************)

unit BasicPlayer;

interface

uses
  StdCtrls, IdBaseComponent, IdComponent, IdUDPBase, IdUDPClient, ExtCtrls,
  math, classes, sysutils, dialogs, windows,
  BasicConst, BasicSettings, BasicType, BasicAssist, BasicObjects;

type
  EFaceDirException = class(Exception);
  ETriangulationException = class(Exception);
  EFailedPositionCalculation = class(Exception);

  TMessageEvent = procedure (Sender: TObject; Text: string) of object;

  TBasicPlayer = class(TidUDPClient)
  private
    { Private de clarations }
    FLog: TStrings;

    FEffort: double;
    FStamina: double;

    FLastSend: integer;
    FgotNewVisualInformation: boolean;

    FCurTime: integer;
    FTeamName: string;
    FPlayNumber: integer;
    FKeeperNr: integer;
    FSide: TSide;
    FViewAngle: TViewAngle;
    FViewQuality: TViewQuality;
    FPlayMode: TPlayMode;
    FRelPlayMode: TRelSide;

    FVersion: double;
    FX: double;
    FY: double;
    FDirection: double;
    FUserData: pointer;

    FCurrentMessageNr: integer;
    FCurrentSeeNr: integer;
    FCurrentHearNr: integer;
    FCurrentSenseNr: integer;
    FCurrentTouchNr: integer;
    FOnRelRefereeMsg: TNotifyEvent;
    FOnSenseMsg: TNotifyEvent;
    FOnSeeMsg: TNotifyEvent;
    FOnErrorMsg: TNotifyEvent;
    FOnUnknownMsg: TNotifyEvent;
    FOnProgress: TNotifyEvent;
    FOnCoachMsg: TMessageEvent;
    FOnPlayerMsg: TMessageEvent;

    FBasicSettings: TBasicParSettings;
    FVisualObjectList: TVisualObjectList;

    procedure SetUserData(const Value: pointer);
    procedure SetKeeperNr(const Value: integer);

    procedure analyzeMessage(cMsg: string);
    procedure SetProgress(const Value: TNotifyEvent);
    procedure SetCoachMsg(const Value: TMessageEvent);
    procedure SetErrorMsg(const Value: TNotifyEvent);
    procedure SetPlayerMsg(const Value: TMessageEvent);
    procedure SetRelRefereeMsg(const Value: TNotifyEvent);
    procedure SetSenseMsg(const Value: TNotifyEvent);
    procedure SetSeeMsg(const Value: TNotifyEvent);
    procedure SetUnknownMsg(const Value: TNotifyEvent);
    procedure CalcAbsFaceDir;
    procedure getAbsPosition;
    procedure setPosition(X, Y: double);
    function GetHalfViewAngle: double;

    procedure calcUnseenImmovableObjects;
    procedure calcMovableObjects;
  public
    { Public declarations }
    property gotNewVisualInformation: boolean read FgotNewVisualInformation;
    property Log: TStrings read FLog write FLog;

    property CurTime: integer read FCurTime;
    property TeamName: string read FTeamName;
    property PlayNumber: integer read FPlayNumber;
    property KeeperNr: integer read FKeeperNr write SetKeeperNr;
    property Side: TSide read FSide;
    property ViewAngle: TViewAngle read FViewAngle;
    property ViewQuality: TViewQuality read FViewQuality;
    property PlayMode: TPlayMode read FPlayMode;
    property RelPlayMode: TRelSide read FRelPlayMode;
    property CurrentMessageNr: integer read FCurrentMessageNr;
    property CurrentSeeNr: integer read FCurrentSeeNr;
    property CurrentHearNr: integer read FCurrentHearNr;
    property CurrentSenseNr: integer read FCurrentSenseNr;
    property CurrentTouchNr: integer read FCurrentTouchNr;
    property halfViewAngle: double read GetHalfViewAngle;

    property X: double read FX;
    property Y: double read FY;
    property Direction: double read FDirection;
    property UserData: pointer read FUserData write SetUserData;

    property OnProgress: TNotifyEvent read FOnProgress write SetProgress;
    property OnRelRefereeMsg: TNotifyEvent read FOnRelRefereeMsg write
      SetRelRefereeMsg;
    property OnCoachMsg: TMessageEvent read FOnCoachMsg write SetCoachMsg;
    property OnPlayerMsg: TMessageEvent read FOnPlayerMsg write SetPlayerMsg;
    property OnSenseMsg: TNotifyEvent read FOnSenseMsg write SetSenseMsg;
    property OnSeeMsg: TNotifyEvent read FOnSeeMsg write SetSeeMsg;
    property OnErrorMsg: TNotifyEvent read FOnErrorMsg write SetErrorMsg;
    property OnUnknownMsg: TNotifyEvent read FOnUnknownMsg write SetUnknownMsg;

    function Send(s: string): boolean;
    procedure mainLoop;
    procedure catchBall(direction: integer);
    procedure changeView(width, quality: integer);
    procedure dash(power: double);
    procedure kick(power, direction: double);
    procedure move(x, y: double);
    procedure moveRel(amount: double);
    procedure say(s: string);
    procedure requestSenseInfo;
    procedure turn(angle: double);
    procedure getScore(var teamscore: integer; var opponentscore: integer);
    function tryStringCommand(command: string): integer;
    procedure showStringCommands(stream: TStream);
    procedure showStatus(stream: TStream);
    procedure joinMatch;
    procedure leaveMatch;

    function Dash2Distance(power: double): double; // estimate distance for dash
    function Kick2Distance(power: double): double; // estimate distance for kick
    function Distance2DashPower(distance: double): double;
      // estimate power for distance
    function Distance2KickPower(distance: double): double;
      // estimate power for distance

    function GetVOList(kindSet: TKindSet; MinConfidence: double = 0.0;
      MaxConfidence: double = 1.0): TVOList;

    constructor Create(team: string = BPLAYER_DEFAULT_PLAYERNAME; host: string =
      BPLAYER_DEFAULT_SERVER_HOST;
      port: integer = BPLAYER_DEFAULT_SERVER_PORT; version: double =
        BPLAYER_DEFAULT_SERVER_VERSION);
    destructor Destroy; override;
  end;

implementation

{ TBasicPlayer }

function TBasicPlayer.Send(s: string): boolean;
var
  nTick: integer;
begin
  Result := False;
  if PlayMode <> pm_illegal then
  begin
    FgotNewVisualInformation := False;
    if Assigned(FLog) then
      FLog.Add('UIT: ' + s);
    nTick := GetTickCount;
    if ((nTick - FLastSend) < BPLAYER_DEF_MINIMUM_SEND_INTERVAL) then
      Sleep(BPLAYER_DEF_MINIMUM_SEND_INTERVAL - max(nTick - FLastSend, 0));
    inherited Send(host, port, s);
    FLastSend := nTick;
    Result := True;
  end;
end;

procedure TBasicPlayer.move(x, y: double);
var
  c: string;
begin
  c := Format('(move %f %f)', [x, y]);
  Send(c);
  FX := x;
  FY := y;
end;

procedure TBasicPlayer.turn(angle: double);
var
  c: string;
begin
  // angle may be -180 to 180
  c := Format('(turn %f)', [angle]);
  Send(c);
  FDirection := FDirection + angle;
  if (FDirection > 180) then
    FDirection := FDirection - 360;
end;

procedure TBasicPlayer.dash(power: double);
var
  c: string;
begin
  // power may be -30 to 100
  c := Format('(dash %f)', [power]);
  Send(c);
end;

procedure TBasicPlayer.kick(power, direction: double);
var
  c: string;
begin
  // power -30 to 100
  // direction -180 to 180
  c := Format('(kick %f %f)', [power, direction]);
  Send(c);
end;

procedure TBasicPlayer.catchBall(direction: integer);
var
  c: string;
begin
  // direction -180 to 180
  c := Format('(catch %d)', [direction]);
  Send(c);
end;

procedure TBasicPlayer.say(s: string);
begin
  Send('(say ' + s + ')');
end;

procedure TBasicPlayer.changeView(width, quality: integer);
var
  c: string;
begin
  // set the width
  c := '(change_view ';
  case width of
    180: c := c + 'wide ';
    90: c := c + 'normal ';
  else
    c := c + 'narrow ';
  end;
  // set the view quality string
  if quality = 0 then
    c := c + 'low)'
  else
    c := c + 'high)';
  Send(c);
end;

procedure TBasicPlayer.requestSenseInfo;
begin
  Send('(sense_body)');
end;

procedure TBasicPlayer.moveRel(amount: double);
var
  newX, newY: double;
begin
  newX := X + cos(Direction * PI / 180);
  newY := Y + sin(Direction * PI / 180);

  // and move to the calculated position
  Move(newX, newY);
end;

constructor TBasicPlayer.Create(team, host: string; port: integer;
  version: double);
begin
  inherited Create(nil);

  FEffort := 1;
  FStamina := 1;
  FgotNewVisualInformation := false;

  FVisualObjectList := TVisualObjectList.Create;
  FBasicSettings := TBasicParSettings.Create;

  FTeamName := team;
  FVersion := version;
  self.Host := host;
  self.Port := port;

  FPlayMode := pm_illegal;
  FRelPlayMode := rside_none;

  FCurrentMessageNr := 0;
  FCurrentSeeNr := 0;
  FCurrentHearNr := 0;
  FCurrentSenseNr := 0;
  FCurrentTouchNr := 0;

  (* Initialize Callback functions *)
  FOnRelRefereeMsg := nil;
  FOnCoachMsg := nil;
  FOnPlayerMsg := nil;
  FOnSenseMsg := nil;
  FOnErrorMsg := nil;
  FOnUnknownMsg := nil;
  FOnProgress := nil;
end;

procedure TBasicPlayer.SetUserData(const Value: pointer);
begin
  FUserData := Value;
end;

destructor TBasicPlayer.Destroy;
begin
  if Active then
  begin
    // close match
    Active := False;
  end;
  FBasicSettings.Free;
  FVisualObjectList.Free;
  inherited;
end;

procedure TBasicPlayer.SetKeeperNr(const Value: integer);
begin
  FKeeperNr := Value;
end;

procedure TBasicPlayer.getScore(var teamscore, opponentscore: integer);
begin
  teamscore := 0;
  opponentscore := 0;
end;

procedure TBasicPlayer.showStatus(stream: TStream);
var
  s: string;
begin
  s := 'Status functions is not supported at this time';
  stream.Write(s, length(s));
end;

procedure TBasicPlayer.showStringCommands(stream: TStream);
var
  s: string;
begin
  s := 'No functions are supported at this time';
  stream.Write(s, length(s));
end;

function TBasicPlayer.tryStringCommand(command: string): integer;
begin
  result := 0;
end;

(*
  Parse incoming information
*)

procedure TBasicPlayer.analyzeMessage(cMsg: string);
var
  cBasic: string;
  cType: string;
  cTemp: string;
  nPos: integer;
  nTemp: integer;
  dTemp: double;
  cObjectInfo: string;
  cObjName: string;
  visObj: TVisualObject;
begin
  if Assigned(FLog) then
    FLog.Add('IN: ' + cMsg);
    
  // First, strip the brackets...
  cMsg := Copy(cMsg, 2, length(cMsg) - 2);

  cBasic := trim(cMsg);
  cType := GetToken(cMsg, 0);

  if cType = 'init' then
  begin // init Side Uniformnumber Playmode
    cTemp := GetToken(cMsg, 1);
    FSide := StrToSide(cTemp);

    FPlayNumber := StrToInt(GetToken(cMsg, 2));
    FPlayMode := StrToPlayMode(GetToken(cMsg, 3));
  end
  else if cType = 'hear' then
  begin // hear Time Direction Message
    FCurTime := GetInteger(cMsg, 1);
    cTemp := GetToken(cMsg, 2);
    if cTemp = 'referee' then
    begin
      // set the playmode
      FPlayMode := StrToPlayMode(GetToken(cMsg, 3));
      if Assigned(FOnRelRefereeMsg) then
        FOnRelRefereeMsg(Self);
    end
    else if cTemp = 'self' then
    begin
      // currently ignore own messages...
    end
    else
    begin
      // it is a say message...
      // is it from the coach or from another player?
      // Assumption: message from coach starts with 'Coach '+TeamName+':'
      cTemp := Copy(cMsg,getTokenPos(cMsg,3),length(cMsg));
      if Copy(cTemp,1,6) = 'Coach ' then
      begin
        if (CompareText(Copy(cTemp,7,Pos(':',cTemp)-7),TeamName) = 0) then
        begin
          if Assigned(FOnCoachMsg) then
            FOnCoachMsg(Self, Copy(cTemp,Pos(':',cTemp)+1,length(cTemp)));
        end
        else
        begin
          // Message from coach from other team.
          if Assigned(FOnPlayerMsg) then
            FOnPlayerMsg(Self, Copy(cTemp,Pos(':',cTemp)+1,length(cTemp)));
        end;
      end
      else
      begin
        // message from other player
        if Assigned(FOnPlayerMsg) then
          FOnPlayerMsg(Self, cTemp);
      end;
    end;
  end
  else if cType = 'sense_body' then
  begin // sense_body Time (sensory information)
    cMsg := RemoveAllBrackets(cMsg);
    FCurTime := GetInteger(cMsg, 1);
    cTemp := GetToken(cMsg, 3);
    if (cTemp = 'high') then
      FViewQuality := vq_high
    else if (cTemp = 'low') then
      FViewQuality := vq_low
    else
      FViewQuality := vq_illegal;

    cTemp := GetToken(cMsg, 4);
    if cTemp = 'narrow' then
      FViewAngle := va_narrow
    else if cTemp = 'normal' then
      FViewAngle := va_normal
    else if cTemp = 'wide' then
      FViewAngle := va_wide
    else
      FViewAngle := va_illegal;

    FStamina := GetDouble(cMsg, 6);
    FEffort := GetDouble(cMsg, 7);

    // amountofspeed: GetDouble(cMsg,9);
    if Assigned(FOnSenseMsg) then
      FOnSenseMsg(Self);
  end
  else if cType = 'error' then
  begin
    if Assigned(FOnErrorMsg) then
      FOnErrorMsg(Self);
  end
  else if cType = 'see' then
  begin // see Time ObjInfo ObjInfo...
    FCurTime := GetInteger(cMsg, 1);
    FVisualObjectList.UpdateDecay(FCurTime);
      // update the decay of the previous data
    nPos := 1;
    repeat
      visObj := nil;
      cObjectInfo := GetBracketPart(cMsg, nPos);
      Inc(nPos);
      // ObjName Distance Direction DistChng DirChng [BodyDir HeadDir]
      nTemp := 2;
      cObjName := RemoveAllBrackets(GetBracketPart(cObjectInfo, nTemp));
      cTemp := GetToken(cObjName, 0);
      if cTemp = 'player' then
      begin
        if GetToken(cObjName, 2) <> '' then
          visObj := TPlayerObject.Create(CurTime, TeamName, Side,
            GetToken(cObjName, 1), GetInteger(cObjName, 2));
      end
      else if cTemp = 'goal' then
      begin
        visObj := TGoalObject.Create(CurTime, Side, StrToSide(GetToken(cObjName,
          1)));
      end
      else if cTemp = 'ball' then
        visObj := TBallObject.Create(CurTime)
      else if cTemp = 'flag' then
      begin
        visObj := TFlagObject.Create(RemoveSpace(Copy(cObjName, 5,
          length(cObjName))), CurTime);
      end
      else if cTemp = 'line' then
      begin
        visObj := TLineObject.Create(CurTime, StrToPosition(GetToken(cObjName,
          1)));
      end;

      if Assigned(visObj) then
      begin
        FgotNewVisualInformation := True;
        // Process distance information before updating
        cObjName := GetLastPart(cObjectInfo);
        dTemp := GetDouble(cObjName, 0); // Could be direction of distance
        cTemp := GetToken(cObjName, 1);
        if (cTemp = '') then
          visObj.Direction := dTemp // 1 parameter -> dTemp is direction
        else
        begin
          visObj.Distance := dTemp; // more parameters -> dTemp is distance
          visObj.Direction := GetDouble(cObjName, 1);
          cTemp := GetToken(cObjName, 2);
          if cTemp <> '' then // more parameters?
          begin
            visObj.DistChng := GetDouble(cObjName, 2);
            visObj.DirChng := GetDouble(cObjName, 3);
            cTemp := GetToken(cObjName, 4);
            if cTemp <> '' then // more parameters?
            begin
              if visObj is TPlayerObject then
              begin
                TPlayerObject(visObj).BodyDir := GetDouble(cObjName, 4);
                cTemp := GetToken(cObjName, 5);
                if cTemp <> '' then // more parameters?
                  TPlayerObject(visObj).HeadDir := GetDouble(cObjName,5);
              end;
            end;
          end;
        end;
        FVisualObjectList.AddObject(visObj);
      end;
    until (cObjectInfo = '');
    try
      getAbsPosition; // Now calculate the absolute position of the player.
      calcUnseenImmovableObjects; // Calculate angle and distance to unseen, immovable objects (known positions).
      calcMovableObjects;         // Calculate players and ball.
    except
    end;
    if Assigned(FOnSeeMsg) then
      FOnSeeMsg(Self);
  end
  else
  begin // unknown
    if Assigned(FOnUnknownMsg) then
      FOnUnknownMsg(Self);
  end;
end;

procedure TBasicPlayer.mainLoop;
var
  c: string;
  Vip: string;
  Vport: integer;
begin
  c := '';
  while Active do
  begin
    vip := host;
    vport := port;
    c := c + ReceiveString(Vip, Vport, 5);
    if vport > 0 then
    begin
      host := Vip;
      port := Vport;
    end;

    //ShowMessage(ip + ' ' + inttostr(port));
    //c := c + ReceiveString(5);
    if c <> '' then
    begin
      while (pos(#10, c) > 0) do
      begin
        // Fsl.Add(copy(c,1,pos(#10,c)-1));
        // Fsl.SaveToFile('c:\1.dat');
        analyzeMessage(copy(c, 1, pos(#10, c) - 1));
        c := copy(c, pos(#10, c) + 1, length(c));
      end;
    end;
    if Assigned(FOnProgress) then
      FOnProgress(Self);
  end;
end;

procedure TBasicPlayer.joinMatch;
begin
  Active := True;
  FPlayMode := pm_before_kick_off;
    // Ugly, but simple. Otherwise the send isn't processed
  DecimalSeparator := '.';
  Send(Format('(init %s (version %f))', [TeamName, FVersion]));
  FPlayMode := pm_illegal;
end;

procedure TBasicPlayer.SetProgress(const Value: TNotifyEvent);
begin
  FOnProgress := Value;
end;

procedure TBasicPlayer.SetCoachMsg(const Value: TMessageEvent);
begin
  FOnCoachMsg := Value;
end;

procedure TBasicPlayer.SetErrorMsg(const Value: TNotifyEvent);
begin
  FOnErrorMsg := Value;
end;

procedure TBasicPlayer.SetPlayerMsg(const Value: TMessageEvent);
begin
  FOnPlayerMsg := Value;
end;

procedure TBasicPlayer.SetRelRefereeMsg(const Value: TNotifyEvent);
begin
  FOnRelRefereeMsg := Value;
end;

procedure TBasicPlayer.SetSenseMsg(const Value: TNotifyEvent);
begin
  FOnSenseMsg := Value;
end;

procedure TBasicPlayer.SetUnknownMsg(const Value: TNotifyEvent);
begin
  FOnUnknownMsg := Value;
end;

function TBasicPlayer.GetVOList(kindSet: TKindSet;
  MinConfidence, MaxConfidence: double): TVOList;
begin
  Result := FVisualObjectList.GetVOList(kindSet, MinConfidence, MaxConfidence);
end;

function TBasicPlayer.Dash2Distance(power: double): double;
begin
  Result := power * PlaySettings.PlayerPowerRate * FEffort;
end;

function TBasicPlayer.Distance2DashPower(distance: double): double;
begin
  Result := distance / PlaySettings.PlayerPowerRate / FEffort;
end;

function TBasicPlayer.Kick2Distance(power: double): double;
begin
  Result := power * PlaySettings.BallPowerRate * FEffort;
end;

function TBasicPlayer.Distance2KickPower(distance: double): double;
begin
  Result := distance / PlaySettings.BallPowerRate / FEffort;
end;

procedure TBasicPlayer.leaveMatch;
begin
  Send('(bye)');
  Sleep(100);
  Active := False;
end;

(*
  Uses all the visible lines to determine the absolute direction.
*)

procedure TBasicPlayer.CalcAbsFaceDir;
var
  voList: TVOList;
  lo: TLineObject;
begin
  voList := GetVOList(TKindLine, 1.0); // Confidence level = 1.0 means visible.
  lo := nil;
  try
    if voList.Count = 0 then
      raise EFaceDirException.Create('Invalid facedir')
    else if voList.Count = 1 then
      lo := voList.Items[0]
    else
    begin
      if (TLineObject(voList.Items[0]).Distance >
        TLineObject(voList.Items[1]).Distance) then
        lo := voList.Items[0]
      else
        lo := voList.Items[1];
    end;
  finally
    voList.Free;
  end;

  // compute absolute position from the line.
  case lo.Position of
    poTop:
      begin
        if lo.Direction >= 0 then
          FDirection := lo.Direction
        else
          FDirection := 180 + lo.Direction;
      end;
    poBottom:
      begin
        if lo.Direction > 0 then
          FDirection := lo.Direction - 180
        else
          FDirection := lo.Direction;
      end;
    poLeft:
      begin
        if lo.Direction > 0 then
          FDirection := 90 + lo.Direction
        else if lo.Direction < 0 then
          FDirection := lo.Direction - 90
        else
          raise ETriangulationException.Create('Special case 1 not handled');
      end;
  else // poRight
      begin
        if lo.Direction > 0 then
          FDirection := lo.Direction - 90
        else if lo.Direction < 0 then
          FDirection := 90 + lo.Direction
        else
          raise ETriangulationException.Create('Special case 2 not handled');
      end;
  end;
end;

(*
  Uses all the visible lines to determine the absolute position.
*)

procedure TBasicPlayer.getAbsPosition;
var
  voList: TVOList;
  i: integer;
  fo: TFlagObject;
  weight: double;
  sum: double;
  weightedSumX: double;
  weightedSumY: double;
  absPosition: TFieldPosition;
begin
  CalcAbsFaceDir;
  voList := GetVOList(TKindFlag, 1.0); // Confidence level = 1.0 means visible.
  try
    sum := 0.0;
    if voList.Count > 0 then
    begin
      weightedSumX := 0.0;
      weightedSumY := 0.0;

      for i := 0 to voList.Count - 1 do
      begin
        fo := voList.Items[i];
        // Calculate position from flag
        absPosition.X := fo.ObjectPosition.Pos.X - (CosAngle(RelToAbs(Direction,
          fo.Direction)) * fo.Distance);
        absPosition.Y := fo.ObjectPosition.Pos.Y - (SinAngle(RelToAbs(Direction,
          fo.Direction)) * fo.Distance);
        // weight position relative to distance
        weight := 1 / fo.Distance;
        weightedSumX := weightedSumX + (weight * absPosition.X);
        weightedSumY := weightedSumY + (weight * absPosition.Y);
        sum := sum + weight;
      end;
      // set best bet for position player
      SetPosition((1 / sum) * weightedSumX, (1 / sum) * weightedSumY);
    end
    else
      raise EFailedPositionCalculation.Create('Failed to calculate position');
  finally
    voList.Free;
  end;
end;

procedure TBasicPlayer.setPosition(X, Y: double);
begin
  FX := X;
  FY := Y;
end;

(*
  Calculate the angle and the distance to every unseen, immovable object.
  The only way to be sure we only update the information that is not seen,
  we check the time of the object with the time of the player.
  You can also use GetVOList(TKindAllVisible,0.0,0.999) but this will fail if
  a DECAY magic number is set to 1.0
*)
procedure TBasicPlayer.calcUnseenImmovableObjects;
var
  voList: TVOList;
  i: integer;
  vo: TVisualObject;
  Xdif, Ydif: double;
begin
  voList := GetVOList(TKindAllVisible);
  try
    for i := 0 to voList.Count - 1 do
    begin
      vo := voList.Items[i];
      if ((vo.TimeCreate < CurTime) and (vo is TImmovableObject)) then
      begin
        Xdif := vo.ObjectPosition.Pos.X - X;
        Ydif := vo.ObjectPosition.Pos.Y - Y;
        // Calculate the absolute angle.
        vo.Direction := (arctan2(Ydif, Xdif) / Pi) * 180;
        // Now make this the relative angle
        vo.Direction := AbsToRel(vo.Direction, Direction);

        vo.Distance := sqrt(Ydif * YDif + Xdif * Xdif);
      end;
    end;
  finally
    voList.Free;
  end;
end;

function TBasicPlayer.GetHalfViewAngle: double;
begin
  case FViewAngle of
    va_narrow: result := 45 / 2;
    va_normal: result := 90 / 2;
    va_wide: result := 180 / 2;
  else
    result := 90 / 2;
  end;
end;

procedure TBasicPlayer.SetSeeMsg(const Value: TNotifyEvent);
begin
  FOnSenseMsg := Value;
end;

(* Calculate the position for every seen object that is movable.
  If the object is seen, just use the angle and distance to calculate the (X,Y)
  If the object is unseen, just use the calculated (X,Y) to guess
  angle and distance
*)
procedure TBasicPlayer.calcMovableObjects;
var
  voList: TVOList;
  i: integer;
  vo: TVisualObject;
  angle : double;
  Xdif, Ydif: double;
begin
  voList := GetVOList(TKindPlayer + TKindBall);
  try
    for i := 0 to voList.Count - 1 do
    begin
      vo := voList.Items[i];
      if (vo.TimeCreate < CurTime) then
      begin // currently unseen
        Xdif := vo.ObjectPosition.Pos.X - X;
        Ydif := vo.ObjectPosition.Pos.Y - Y;
        // Calculate the absolute angle.
        vo.Direction := (arctan2(Ydif, Xdif) / Pi) * 180;
        // Now make this the relative angle
        vo.Direction := AbsToRel(vo.Direction, Direction);
        vo.Distance := sqrt(Ydif * YDif + Xdif * Xdif);
      end
      else
      begin // currently seen
        angle := RelToAbs(Direction, vo.Direction);
        vo.SetFieldPosition(X + (cos((angle/180) * pi) * vo.Distance),
          Y + (sin((angle/180) * pi) * vo.Distance));
      end;
    end;
  finally
    voList.Free;
  end;
end;

end.
